#include "scene.h"
#include "m33.h"
#include "ppc.h"

#include <iostream>

using namespace std;

Scene *scene;

Scene::Scene() {

  sf = 0.8f;
  sh = 1.0f;
  lightDirv = V3(0.0f, 0.0f, 1.0f);

  cgi = new CGInterface();
  soi = new ShaderOneInterface();

  gui = new GUI();
  gui->show();

  int u0 = 20;
  int v0 = 50;
  int sci = 1;
  int w = sci*480;
  int h = sci*360;
  fb = new FrameBuffer(u0, v0, w, h);
  fb->label("SW Framebuffer");
  fb->show();
  
  hwfb = new FrameBuffer(u0+fb->w+20, v0, w, h);
  hwfb->label("HW Framebuffer");
  hwfb->isHW = true;
  hwfb->show();

  gui->uiw->position(2*fb->w+u0 + 2*20, v0);
  float hfov = 45.0f;
  ppc = new PPC(hfov, w, h);
  tmsN = 10;
  tms = new TMesh[tmsN];
  V3 center(0.0f, 0.0f, -175.0f);
  float sl = 30.0f;
  tms[0].SetCube(center, sl);
  tms[0].enabled = false;
  tms[1].Load("geometry/teapot1K.bin");
  AABB aabb = tms[1].GetAABB();
  float size0 = (aabb.corners[1]-aabb.corners[0]).Length();
  V3 tcenter = tms[1].GetCenter();
  tms[1].Translate(tcenter*-1.0f+center);
  float size1 = 170.0f;
  tms[1].ScaleAboutCenter(size1/size0);
  tms[1].renderWF = false;

#if 0
  tms[9].Load("geometry/auditorium.bin");
  aabb = tms[9].GetAABB();
  size0 = (aabb.corners[1]-aabb.corners[0]).Length();
  tcenter = tms[9].GetCenter();
  tms[9].Translate(tcenter*-1.0f+center);
  size1 = 170.0f;
  tms[9].ScaleAboutCenter(size1/size0);
  tms[9].renderWF = true;
#else
  tmsN = 9;
#endif

  toVis = PPC(45.0f, ppc->w, ppc->h);
  toVis.C = V3(0.0f, 0.0f, -100.0f);

  tms[0].enabled = 
  tms[2].enabled = false;

  tms[3].SetRectangle(600.0f, 60.0f);
  tms[3].Translate(V3(0.0f, 0.0f, -200.0f));
  tms[3].cols[2] = 
  tms[3].cols[3] = V3(0.0f, 1.0f, 0.0f);

  tms[4].SetRectangle(600.0f, 60.0f);
  tms[4].Translate(V3(0.0f, 70.0f, -200.0f));
  tms[4].cols[2] = 
  tms[4].cols[3] = V3(0.0f, 1.0f, 0.0f);
  tms[4].msi = false;

  tms[5].SetRectangle(10.0f, 10.0f);
  tms[5].Translate(V3(0.0f, 35.0f, -200.0f));
  tms[5].cols[0] = 
  tms[5].cols[1] = 
  tms[5].cols[2] = 
  tms[5].cols[3] = V3(1.0f, 0.0f, 0.0f);

  tms[3].enabled = 
  tms[4].enabled = 
  tms[5].enabled = false;

  FrameBuffer *tmap = new FrameBuffer(u0, fb->h+v0+40, 1024, 1024);
  unsigned int cols[2] = {0xFF000000, 0xFFFFFFFF};
  tmap->SetAsChecker(1, cols);
//  tmap->show();
  tmap->label("tmap");
  tms[6].SetRectangle(600.0f, 600.0f, tmap);
  tms[6].Translate(V3(0.0f, 0.0f, -100.0f));
//  tms[6].msi = false;
  tms[6].enabled = false;

  tms[7].SetRectangle(100.0f, 80.0f);
  tms[7].RotateAboutArbitraryAxis(V3(0.0f, 0.0f, 0.0f), V3(1.0f, 0.0f, 0.0f), 90.0f);
  tms[7].Translate(V3(0.0f, 0.0f, -200.0f));
  tms[7].cols[0] = 
  tms[7].cols[1] = 
  tms[7].cols[2] = 
  tms[7].cols[3] = V3(1.0f, 0.0f, 0.0f);
  tms[7].enabled = false;

  tms[8].SetRectangle(10.0f, 20.0f);
  tms[8].Translate(V3(0.0f, 10.0f, -200.0f));
  tms[8].cols[0] = 
  tms[8].cols[1] = 
  tms[8].cols[2] = 
  tms[8].cols[3] = V3(0.0f, 1.0f, 0.0f);
  tms[8].enabled = false;

  refPPC = 0;
  refFB = 0;

//  LoadView0();

  Render();

}

void Scene::FrameSetup() {

  fb->Set(0xFF7F0000);
  fb->SetZB(0.0f);

}

void Scene::Render() {

  if (hwfb) {
    hwfb->redraw();
  }

  return;

//  cerr << "INFO: render SW" << endl;

  // SW rendering
  FrameSetup();
  for (int tmi = 0; tmi < tmsN; tmi++) {
    if (!tms[tmi].enabled)
      continue;
    if (tms[tmi].renderWF) {
      tms[tmi].RenderWF(ppc, fb);
    }
    else {
      tms[tmi].Render(ppc, fb);
    }
  }

//  toVis.Visualize(ppc, fb, 10.0f, 8);

  fb->redraw();

}

void Scene::DBG() {

  {
    int stepsN = 100;
    for (int i = 0; i < stepsN-3; i++) {
      sh = 1.0f*(float) i / (float) (stepsN-1);
      Render();
      Fl::check();
    }
    return;
  }

  {
    V3 l0 = V3(0.0f, 0.0f, 1.0f);
    V3 l1 = V3(0.0f, 1.0f, 1.0f);
    l1 = l1.Normalized();
    int stepsN = 100;
    for (int i = 0; i < stepsN; i++) {
      float frac = (float) i / (float) (stepsN-1);
      lightDirv =  (l0 + (l1-l0)*frac).Normalized();
      Render();
      Fl::check();
    }
    return;
  }

  ppc->zNear = 144.0f;
//  cerr << ppc->zNear << endl;
  Render();
  return;

  ProjectiveTextureMappingSetup();
  return;

  tms[1].enabled = false;
  tms[2].SetFromFB(fb, ppc);
  Render();
  return;

  float visF = 10.0f;
  for (int si = 0; si < 360; si++) {
    toVis.Pan(1.0f);
    toVis.Visualize(ppc, fb, visF, 8);
    Render();
    Fl::check();
  }
  return;

  V3 center = tms[1].GetCenter();
  V3 newC(250.0f, 100.0f, 60.0f);
  V3 newVD = (center-newC).Normalized();
  V3 vinUP(0.0f, 1.0f, 0.0f);
  PPC ppc1;
  ppc->PositionAndOrient(newC, newVD, vinUP, ppc1);
  PPC ppc0(*ppc);
  int sN = 100;
  for (int si = 0; si < sN; si++) {
    float frac = (float) si / (float) sN;
    *ppc = ppc->Interpolate(&ppc0, &ppc1, frac);
    Render();
    Fl::check();
  }
  return;

  float tstep = 2.0f;
  ppc->TranslateLeft(tstep);
  Render();
  return;

  int stepsN = 360;
  V3 center1 = tms[0].GetCenter();
  for (int si = 0; si < stepsN; si++) {
    tms[0].RotateAboutArbitraryAxis(center1, V3(0.0f, 1.0f, 0.0f), 1.0f);
    Render();
    Fl::check();
  }
  return;

  float hfov = 45.0f;
  PPC ppc(hfov, fb->w, fb->h);
  
  float sl = 30.0f;
  V3 cvs[8];
  float nearz = -sl*2.0f;
  cvs[0] = V3(-sl/2.0f, sl/2.0f, nearz);
  cvs[1] = V3(-sl/2.0f, -sl/2.0f, nearz);
  cvs[2] = V3(sl/2.0f, -sl/2.0f, nearz);
  cvs[3] = V3(sl/2.0f, sl/2.0f, nearz);
  cvs[4] = V3(-sl/2.0f, sl/2.0f, nearz-sl);
  cvs[5] = V3(-sl/2.0f, -sl/2.0f, nearz-sl);
  cvs[6] = V3(sl/2.0f, -sl/2.0f, nearz-sl);
  cvs[7] = V3(sl/2.0f, sl/2.0f, nearz-sl);
  V3 projPs[8];
  V3 cols[8];
  cols[0] = cols[1] = cols[2] = cols[3] = V3(1.0f, 0.0f, 0.0f);
  cols[4] = cols[5] = cols[6] = cols[7] = V3(0.0f, 0.0f, 0.0f);
  for (int i = 0; i < 8; i++) {
    ppc.Project(cvs[i], projPs[i]);
  }
  for (int i = 0; i < 4; i++) {
    int _i = (i+1)%4;
    fb->DrawSegment(projPs[i], projPs[_i], cols[i], cols[_i]);
    fb->DrawSegment(projPs[4+i], projPs[4+_i], cols[4+i], cols[4+_i]);
    fb->DrawSegment(projPs[i], projPs[4+i], cols[i], cols[4+i]);
  }
  Render();
  return;


  return;
  V3 P(0.0f, 0.0f, -1.0f);
  V3 projP;
  bool projOK = ppc.Project(P, projP);
  return;

  M33 m;
  m[0] = V3(1.0f, 3.0f, -1.0f);
  m[1] = V3(10.0f, -31.0f, -11.0f);
  m[2] = V3(11.0f, 13.0f, 12.0f);
  M33 mi  = m.Invert();
  cerr << mi*m << endl << endl << m*mi;
  return;

  fb->Set(0xFF308080);
  Render();
  Fl::check();
  float scf = 10.5f;
  fb->Brighten(scf);
  Render();
  return;

  V3 P0(222.5f, 10.5f, 0.0f);
  V3 P1(122.5f, 410.5f, 0.0f);

  V3 c0;
  c0.SetFromColor(0xFF0000FF);
  V3 c1;
  c1.SetFromColor(0xFF00FF00);

  fb->DrawSegment(P0, P1, c0, c1);
  Render();
  return;

  unsigned int colss[2] = {0xFF000000, 0xFFFFFFFF};
  fb->SetAsChecker(32, colss);
  Render();
  return;

  for (int u = 0; u < fb->w; u++) {
    fb->Set(u, fb->h/2, 0xFF000000);
  }
  Render();
  return;

  cerr << "INFO: DBG" << endl;
  fb->Set(0xFF00FFFF);
  Render();

}

void Scene::SaveView0() {

  ppc->Save("mydbg/view0.txt");

}

void Scene::LoadView0() {

  ppc->Load("mydbg/view0.txt");
  Render();

}

void Scene::GoToView0() {

  PPC nppc;
  nppc.Load("mydbg/view0.txt");
  GoToView(&nppc);

}

void Scene::SaveView1() {

  ppc->Save("mydbg/view1.txt");

}

void Scene::LoadView1() {

  ppc->Load("mydbg/view1.txt");
  Render();

}

void Scene::GoToView1() {

  PPC nppc;
  nppc.Load("mydbg/view1.txt");
  GoToView(&nppc);

}


void Scene::GoToView(PPC *nppc) {

  PPC oppc(*ppc);
  int n = 100;
  for (int i = 0; i < n; i++) {
    float frac = (float) i / (float) (n-1);
    *ppc = ppc->Interpolate(&oppc, nppc, frac);
    Render();
    Fl::check();
  }



}


void Scene::ProjectiveTextureMappingSetup() {

  if (refPPC) {
    delete refPPC;
    delete refFB;
  }


  refPPC = new PPC(*ppc);

  refFB = new FrameBuffer(20, fb->h+50, fb->w, fb->h);
  refFB->CopyFrom(fb);
  refFB->label("Reference Image");
  refFB->show();

  tms[7].projPPC = refPPC;
  tms[7].projTM = refFB;
  tms[8].projPPC = refPPC;
  tms[8].projTM = refFB;

}


void Scene::FrameSetupHW() {

  // OpenGL setup
  glEnable(GL_DEPTH_TEST);
//  glEnable(GL_CULL_FACE);

  // frame setup
  glClearColor(0.0f, 0.0f, 0.5f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | 
    GL_DEPTH_BUFFER_BIT);

  // set view
  // set intrinsics
  ppc->SetIntrinsicsHW();
  // set extrinsics
  ppc->SetExtrinsicsHW();


}

void Scene::RenderHW() {

  /// OpenGL frame setup
  FrameSetupHW();

  // render geometry: issue geometry for rendering
  for (int tmi = 0; tmi < tmsN; tmi++) {
    if (!tms[tmi].enabled)
      continue;
    tms[tmi].RenderHW();
  }

}

void Scene::RenderGPU() {

  if (cgi->needInit) {
    cgi->PerSessionInit();
    soi->PerSessionInit(cgi);
  }

  FrameSetupHW();

  ppc->RenderImageFrameGL();

  // per frame parameter setting and enabling programs
  soi->PerFrameInit();
  soi->BindPrograms();
  cgi->EnableProfiles();

  tms[1].RenderHW();

  soi->PerFrameDisable();
  cgi->DisableProfiles();

}

